/*
	SetSubsys v1.00, (c)2000 Jeremy Collake
	http://www.collakesoftware.com
	mailto:collake@charter.net
	--------------------------------------------------------------------------------

	This simple, console mode utility will display, and optionally set, the subsystem
	of win32 portable executables.
	
	Usage:
		
		Run SetSubsys.exe with no parameters for help.

*/

#include<stdio.h>
#include<conio.h>
#include<windows.h>
#include<winuser.h>
#include<imagehlp.h>
#include<winnt.h>

#define SUBSYSTEM_NOT_SET 13
#define ERROR_USAGE 1
#define ERROR_NOFILES 2
#define ERROR_SET_ERROR 3
#define ERROR_NOT_PE    4
#define ERROR_MAP_FAILURE 5
#define ERROR_MAPVIEW_FAILURE 6
#define ERROR_UNICODE_FAILURE 7
#define ERROR_OPEN_FAILURE 8

#define MAX_SUBSYSTEM 13

char cSwitchArray[] =
{
	'C',
	'W',
	'O',
	'P',
	'N',
	'9',
	'E',
	'A',
	'B',
	'R',
	'U',
	0
};

unsigned short nSubsystemSwitchParallel[] =
{
	IMAGE_SUBSYSTEM_WINDOWS_CUI,
	IMAGE_SUBSYSTEM_WINDOWS_GUI,
	IMAGE_SUBSYSTEM_OS2_CUI,
	IMAGE_SUBSYSTEM_POSIX_CUI,
	IMAGE_SUBSYSTEM_NATIVE,
	IMAGE_SUBSYSTEM_NATIVE_WINDOWS,
	IMAGE_SUBSYSTEM_WINDOWS_CE_GUI,
	IMAGE_SUBSYSTEM_EFI_APPLICATION,
	IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER,
	IMAGE_SUBSYSTEM_EFI_RUNTIME_DRIVER,
	IMAGE_SUBSYSTEM_UNKNOWN
};

char *pszSubsystemNamesArray[] =
{
	"Unknown",
	"NT Native",
	"Windows GUI",
	"Windows Console",
	"Unknown",
	"OS/2",
	"Unknown"
	"POSIX",
	"Win9x Native",
	"WindowsCE",
	"EFI Application",
	"EFI Boot Service Driver",
	"EFI Runtime Driver"
};

bool DoesDirectoryExist(char *pDirName) {
	unsigned int nAttributes=0;
	if((nAttributes=GetFileAttributes(pDirName))==-1)
		return false;
	if((nAttributes&FILE_ATTRIBUTE_DIRECTORY)!=0)
		return true;
	else return false;
}

int SetNewSubsystem(char *szFile, unsigned short nTargetSubsystem)
{
		HANDLE hFile;
		HANDLE hMapping;
		unsigned char *pView;
		PIMAGE_NT_HEADERS pNTHeader;
		PIMAGE_DOS_HEADER pDOS;		

		DWORD nFileSize;
						
		hFile=CreateFile(szFile,GENERIC_READ|GENERIC_WRITE,FILE_SHARE_READ,0,OPEN_EXISTING,0,0);
		if(hFile==INVALID_HANDLE_VALUE)
		{
			return ERROR_OPEN_FAILURE;
		}
		nFileSize=GetFileSize(hFile,NULL);
		hMapping=CreateFileMapping(hFile,0,PAGE_READWRITE,0,0,NULL);
		if(!hMapping)
		{
			CloseHandle(hFile);
			return ERROR_MAP_FAILURE;
		}		
		pView=(unsigned char *)MapViewOfFile(hMapping,FILE_MAP_ALL_ACCESS,0,0,0);
		if(!pView)
		{
			CloseHandle(hMapping);
			CloseHandle(hFile);
			return ERROR_MAPVIEW_FAILURE;
		}		
		pDOS=(PIMAGE_DOS_HEADER)pView;
		pNTHeader=(PIMAGE_NT_HEADERS)(pDOS->e_lfanew+(DWORD)pDOS);		
		if(pDOS->e_magic!=IMAGE_DOS_SIGNATURE
			||
			IsBadReadPtr((void *)pNTHeader,sizeof(_IMAGE_OPTIONAL_HEADER))
			||
			pNTHeader->Signature!=IMAGE_NT_SIGNATURE)
		{			
			UnmapViewOfFile((LPCVOID)pView);
			CloseHandle(hMapping);
			CloseHandle(hFile);
			return ERROR_NOT_PE;
		}				
		printf("\n Current subsystem:   %s", 
			pNTHeader->OptionalHeader.Subsystem>12?"Unknown":pszSubsystemNamesArray[pNTHeader->OptionalHeader.Subsystem]);
		// if subsystem not specified by user, do not set new subsystem		
		if(nTargetSubsystem!=SUBSYSTEM_NOT_SET)
		{
			pNTHeader->OptionalHeader.Subsystem=nTargetSubsystem;			
			printf("\n New subsystem:       %s", 
				pNTHeader->OptionalHeader.Subsystem>12?"Unknown":pszSubsystemNamesArray[pNTHeader->OptionalHeader.Subsystem]);
		}		
		UnmapViewOfFile((LPCVOID)pView);
		CloseHandle(hMapping);
		CloseHandle(hFile);			
		return ERROR_SUCCESS;		
}


int main(int argc, char **argv)
{
	WIN32_FIND_DATA FindData;
	HANDLE findHandle;
	char szFilespec[MAX_PATH*2];
	char szPath[MAX_PATH*2];
	char szFile[MAX_PATH*2];
	unsigned short nTargetSubsystem=SUBSYSTEM_NOT_SET;

	printf("\n SetSubSys v1.00, (c)2000 Jeremy Collake");
	printf("\n http://www.collakesoftware.com");
	printf("\n mailto:collake@charter.net");
	printf("\n ----------------------------------------------------------------------");
	if(argc<2)
	{
showhelp:
		printf("\n Usage: SetSubSys filespec [/C|/W|/O|/P|/N|/9|/E|/A|/B|/R|/U]");
		printf("\n   /C = Windows Console");
		printf("\n   /W = Windows GUI");
		printf("\n   /O = OS/2");
		printf("\n   /P = Posix");
		printf("\n   /N = Native NT");
		printf("\n   /9 = Native 9x");
		printf("\n   /E = WindowsCE");
		printf("\n   /A = EFI Application");
		printf("\n   /B = EFI Boot Service Driver");
		printf("\n   /R = EFI Runtime Driver");
		printf("\n   /U = Unknown");
		printf("\n\n If a new subsystem is not specified, the current subystem is displayed.");
		return ERROR_USAGE;
	}		
	for(int nN=1;nN<argc;nN++)
	{
		if(argv[nN][0]=='/')
		{
			if(toupper(argv[nN][1])=='?' || toupper(argv[nN][1])=='H')
			{
				goto showhelp;
			}
			char cU=toupper(argv[nN][1]);
			for(unsigned int nN=0;cSwitchArray[nN];nN++)
			{
				if(cU==cSwitchArray[nN])
				{
					nTargetSubsystem=nSubsystemSwitchParallel[nN];
					break;
				}
			}		
			if(!cSwitchArray[nN])			
			{
				printf("\n Unrecognized switch: %s.", argv[nN]);
			}
		}
		else
		{			
			strcpy(szPath,argv[nN]);
			strcpy(szFilespec,argv[nN]);	
		}
	}
	if(!DoesDirectoryExist(szPath))
	{				
		if((strstr(szPath,"\\")!=NULL || strstr(szPath,"/")!=NULL) && strlen(szPath))
		{		
			for(int nI=strlen(szPath)-1;
				szPath[nI]!=0 && szPath[nI]!='\\' && szPath[nI]!='/' && nI>0;
				nI--);		
			szPath[nI]=0;
		}	
		else
		{
			strcpy((char *)&szPath,".");
		}
	}
	else
	{
		if(szPath[strlen(szPath)-1]=='\\')
		{
			szPath[strlen(szPath)-1]=0;
		}
		strcpy(szFilespec,szPath);
		strcat(szFilespec,"\\*.*");
	}
	findHandle=NULL;
	while(1)
	{		
		if(!findHandle)
		{
			findHandle=FindFirstFile(szFilespec,&FindData);
			if(findHandle==INVALID_HANDLE_VALUE)
			{	
				printf("\n No files found!\n");
				return ERROR_NOFILES;
			}	
		}
		else
		{
			if(!FindNextFile(findHandle,&FindData))
			{
				break;
			}				
		}
		if((FindData.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY)!=0)
		{
			continue;
		}
		strcpy(szFile,szPath);
		strcat(szFile,"\\");
		strcat(szFile,FindData.cFileName);
		printf("\n %s = ",szFile);

		switch(SetNewSubsystem(szFile,nTargetSubsystem))
		{
		case ERROR_SUCCESS:			
			break;
		case ERROR_OPEN_FAILURE:			
			printf("\n Error opening file!");
			break;
		case ERROR_MAP_FAILURE:
			printf("\n Error mapping file!");
			break;
		case ERROR_MAPVIEW_FAILURE:
			printf("\n Error creating view!");
			break;
		case ERROR_UNICODE_FAILURE:
			printf("\n Error converting filename to Unicode!");
			break;
		default:
			printf("\n Unknown error!");
			break;
		}					

	}
	printf("\n\n Done!");
	return ERROR_SUCCESS;
}
